#include "Headers.h"
#include "HostFlow.h"
#include "RouteHost.h"
#include "PacketDummy.h"
#include "Console.h"
#include "Debug.h"

CHostFlow::CHostFlow(
	__uint32		id,
	CSimHandler*	sim,
	CAddress		address,
	CInfo*			info
	) : CNode(id, sim, address, 1, &CRouteHost::route, info)
{
	// Delegates
	this->delegateRecv = new Delegate4<CHostFlow, void, __uint32, CPacketIp*, CPacketUdp*, CPacket*>(this, &CHostFlow::Recv);
	assert_mem(this->delegateRecv);

	this->delegateConnectionRecv = new Delegate2<CHostFlow, void, CConnectionReceiver*, CPacket*>(this, &CHostFlow::ConnectionRecv);
	assert_mem(this->delegateConnectionRecv);

	this->delegateAccept = new Delegate2<CHostFlow, bool, CAddress, CPacket*>(this, &CHostFlow::ConnectionAccept);
	assert_mem(this->delegateAccept);

	this->delegateAccepted = new Delegate1<CHostFlow, void, CConnectionSender*>(this, &CHostFlow::ConnectionAccepted);
	assert_mem(this->delegateAccepted);

	this->delegateOpen = new Delegate2<CHostFlow, void, CConnection*, CConnection::EOpenResult>(this, &CHostFlow::ConnectionOpen);
	assert_mem(this->delegateOpen);

	this->delegateClose = new Delegate2<CHostFlow, void, CConnection*, CConnection::ECloseResult>(this, &CHostFlow::ConnectionClose);
	assert_mem(this->delegateClose);

	// UDP layer
	(*this->layerUdp->EventRecv()) += this->delegateRecv;

	// Info
	this->streamRate = 500000;
	this->packetRate = 25;
	this->packetInterval = 1.0/this->packetRate; 
	this->packetSize = (__uint32)ceil(this->streamRate / this->packetRate);
	this->connectionsMax = 10;
	this->port = 0;
	this->packetIndex[0] = 0;
	this->packetIndex[1] = 0;
	this->packets = 0;

	// Connection layer
	this->connectionLayer = new CConnectionLayer(
		this->sim,
		this->port,								// port
		this->connectionsMax,					// connections max
		this->layerUdp->DelegateSend(),
		this->delegateConnectionRecv,
		this->delegateAccept,
		this->delegateAccepted,
		this->packetSize						// sender segment size
		);
	assert_mem(this->connectionLayer);

	// Connections
	this->receiver = NULL;
	this->senders[0] = NULL;
	this->senders[1] = NULL;
	this->senderIndex = 0;

	// Timer
	this->timers[0] = new CTimer<CHostFlow>(this->sim, this, &CHostFlow::Timer);
	assert_mem(this->timers[0]);
	this->timers[1] = new CTimer<CHostFlow>(this->sim, this, &CHostFlow::Timer);
	assert_mem(this->timers[1]);

	// Data
	char fileName[256];
	sprintf(fileName, "host_%u_sender.dat", this->id);
	if(FILE_OPEN(this->fileSender, fileName, "w")) { printf("\nCannot create file \'%s\'.", fileName);  exit(-1); }

	sprintf(fileName, "host_%u_receiver.dat", this->id);
	if(FILE_OPEN(this->fileReceiver, fileName, "w")) { printf("\nCannot create file \'%s\'.", fileName);  exit(-1); }

	// Delegates
	this->delegateSenderSend = new Delegate2<CHostFlow, void, CConnectionSender*, CPacket*>(this, &CHostFlow::SenderSend);
	assert_mem(this->delegateSenderSend);
}

CHostFlow::~CHostFlow()
{
	// Connection layer
	delete this->connectionLayer;

	// Delegates
	delete this->delegateRecv;
	delete this->delegateConnectionRecv;
	delete this->delegateAccept;
	delete this->delegateAccepted;
	delete this->delegateOpen;
	delete this->delegateClose;

	// Timer
	delete this->timers[0];
	delete this->timers[1];

	// Data
	fclose(this->fileSender);
	fclose(this->fileReceiver);

	// Delegates
	delete this->delegateSenderSend;
}

void CHostFlow::Connect(CAddress dst)
{
	assert(NULL == this->receiver);

#ifdef LOG
	printf("\nT = %8.3lf\tPEER %u CONNECT %u:%u : ", this->sim->Time(), this->address.Address(), dst.Address(), 100);
#endif

	// Create receiver connection
	switch(this->connectionLayer->Create(dst, this->port, &this->receiver))
	{
	case CConnectionLayer::SUCCESS:
#ifdef LOG
		CConsole::SetColor(CConsole::LIGHT_GREEN);
		printf("SUCCESS");
		CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif

		// Add events to the connection
		(*this->receiver->EventOpen()) += this->delegateOpen;
		(*this->receiver->EventClose()) += this->delegateClose;

		// Open the connection
		this->receiver->Open();

		break;
	case CConnectionLayer::FAIL_MAX_CONNECTIONS_REACHED:
#ifdef LOG
		CConsole::SetColor(CConsole::LIGHT_RED);
		printf("FAIL");
		CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif
		break;
	}

}

void CHostFlow::DisconnectSender(__uint8 sender)
{
#ifdef LOG
	printf("\nT = %8.3lf\tPEER %u DISCONNECT SENDER %s", this->sim->Time(), this->address.Address(), this->senders[sender]->ToString());
#endif
	this->senders[sender]->Close();
}

void CHostFlow::DisconnectReceiver()
{
#ifdef LOG
	printf("\nT = %8.3lf\tPEER %u DISCONNECT RECEIVER %s", this->sim->Time(), this->address.Address(), this->receiver->ToString());
#endif
	this->receiver->Close();
}

void CHostFlow::Finalize()
{
	// Call the base class finalizer
	CNode::Finalize();
}

void CHostFlow::ConnectionRecv(CConnectionReceiver* connection, CPacket* packet)
{
#ifdef LOG_VERBOSE
	CConsole::SetColor(CConsole::LIGHT_CYAN);
	printf("\nT = %8.3lf\tPEER %u RECV on %s %u bits (%s) : %u : ", this->sim->Time(), this->address.Address(), connection->ToString(), packet->Size(), packet->ToString(), this->packets++);
	CConsole::SetColor(CConsole::LIGHT_MAGENTA);
	printf(" : %lf bps", this->Link(0)->MeterUtil(1));
	CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif
	// Data
	fprintf(this->fileReceiver, "%.9lf %u %lf %u\n",
		this->sim->Time(),
		packet->Size(),
		this->Link(0)->MeterUtil(1),
		this->Link(0)->MeterQueue(1, 1));
}

bool CHostFlow::ConnectionAccept(CAddress src, CPacket* packet)
{
#ifdef LOG
	printf("\nT = %8.3lf\tPEER %u ACCEPT %u", this->sim->Time(), this->address.Address(), src.Address());
#endif
	return true;
}

void CHostFlow::ConnectionAccepted(CConnectionSender* connection)
{
#ifdef LOG
	printf("\nT = %8.3lf\tPEER %u ACCEPTED %s", this->sim->Time(), this->address.Address(), connection->ToString());
#endif

	assert(this->senderIndex < 2);
	// Set the sender
	this->senders[this->senderIndex] = connection;

	// Set events
	(*this->senders[this->senderIndex]->EventOpen()) += this->delegateOpen;
	(*this->senders[this->senderIndex]->EventClose()) += this->delegateClose;
	(*this->senders[this->senderIndex]->EventSend()) += this->delegateSenderSend;
	this->senderIndex++;
}

void CHostFlow::ConnectionOpen(CConnection* connection, CConnection::EOpenResult result)
{
#ifdef LOG
	printf("\nT = %8.3lf\tPEER %u OPENED %s : ", this->sim->Time(), this->address.Address(), connection->ToString());
#endif

	switch(result)
	{
	case CConnection::OPEN_SUCCESS:
#ifdef LOG
		CConsole::SetColor(CConsole::LIGHT_GREEN);
		printf("SUCCESS");
		CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif
		break;
	case CConnection::OPEN_CANCELED:
#ifdef LOG
		CConsole::SetColor(CConsole::LIGHT_YELLOW);
		printf("CANCELED");
		CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif
		break;
	case CConnection::OPEN_FAIL_TIMEOUT:
#ifdef LOG
		CConsole::SetColor(CConsole::LIGHT_RED);
		printf("FAIL_TIMEOUT");
		CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif
		break;
	case CConnection::OPEN_FAIL_REMOTE:
#ifdef LOG
		CConsole::SetColor(CConsole::LIGHT_RED);
		printf("FAIL_REMOTE");
		CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif
		break;
	}
}

void CHostFlow::ConnectionClose(CConnection* connection, CConnection::ECloseResult result)
{
#ifdef LOG
	printf("\nT = %8.3lf\tPEER %u CLOSED %s : ", this->sim->Time(), this->address.Address(), connection->ToString());
#endif

	switch(result)
	{
	case CConnection::CLOSE_CONFIRMED:
#ifdef LOG
		CConsole::SetColor(CConsole::LIGHT_YELLOW);
		printf("CONFIRMED");
		CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif
		break;
	case CConnection::CLOSE_TIMEOUT:
#ifdef LOG
		CConsole::SetColor(CConsole::LIGHT_RED);
		printf("TIMEOUT");
		CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif
		break;
	case CConnection::CLOSE_COMPLETE:
#ifdef LOG
		CConsole::SetColor(CConsole::LIGHT_GREEN);
		printf("COMPLETE");
		CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif

		printf("\n\nPEER %u\tuplink: %llu\tdownlink: %llu", this->address.Address(),
			this->Link(0)->StatBits(0, 0) + this->Link(0)->StatBits(1, 0),
			this->Link(0)->StatBits(0, 1) + this->Link(0)->StatBits(1, 1)
			);

		break;
	}
}

void CHostFlow::Recv(__uint32 entry, CPacketIp* ip, CPacketUdp* udp, CPacket* packet)
{
	assert(packet);
	
	switch(packet->Type())
	{
	// Pass all connection packets to the connection layer
	case PACKET_TYPE_CONNECTION: this->connectionLayer->Recv(ip->Src(), udp->Src(), udp->Dst(), type_cast<CPacketConnection*>(packet)); break;
	default: assert(0);
	}
}

void CHostFlow::StartFlow(__uint8 sender)
{
#ifdef LOG
	printf("\nT = %8.3lf\tPEER %u START FLOW %s", this->sim->Time(), this->address.Address(), senders[sender]->ToString());
#endif

	// Start sending packets
	this->Timer((CTimerInfo*)sender);
}

void CHostFlow::StopFlow(__uint8 sender)
{
#ifdef LOG
	printf("\nT = %8.3lf\tPEER %u STOP FLOW %s", this->sim->Time(), this->address.Address(), senders[sender]->ToString());
#endif

	// Stop sending packets
	if(this->timers[sender]->IsSet()) this->timers[sender]->Cancel();
}

void CHostFlow::Timer(CTimerInfo* info)
{
	__uint32 sender = 0; //(__uint32)info;

	// Transmit one packet using the connection
	CPacketDummy* packet = new CPacketDummy(this->packetSize, this->packetIndex[0]++);

#ifdef LOG_VERBOSE
	printf("\nT = %8.3lf\tPEER %u SEND on %s %u bits (%s) : %u : ", this->sim->Time(), this->address.Address(), senders[sender]->ToString(), packet->Size(), packet->ToString(), this->packets++);
#endif

	CConnectionSender::EResult result = this->senders[sender]->Send(packet);

	switch(result)
	{
	case CConnectionSender::SEND_SUCCESS:
#ifdef LOG_VERBOSE
		CConsole::SetColor(CConsole::LIGHT_GREEN);
		printf("\nSUCCESS");
		CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif
		break;
	case CConnectionSender::SEND_FAIL_NOT_OPEN:
#ifdef LOG_VERBOSE
		CConsole::SetColor(CConsole::LIGHT_RED);
		printf("\nFAIL_NOT_OPEN");
		CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif
		break;
	}

#ifdef LOG_VERBOSE
	CConsole::SetColor(CConsole::LIGHT_MAGENTA);
	printf(" : %lf bps", this->Link(0)->MeterUtil(0));
	CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif
	// Schedule next timer
	this->timers[sender]->SetAfter(this->packetInterval, (CTimerInfo*)sender);
}

void CHostFlow::SenderSend(CConnectionSender* sender, CPacket* packet)
{
	// Data
	fprintf(this->fileSender, "%.9lf %u %.9lf %u %.9lf %.9lf %.9lf\n",
		this->sim->Time(),
		packet->Size(),
		this->Link(0)->MeterUtil(0),
		this->Link(0)->MeterQueue(1, 0),
		sender->StatFlowRtt(),
		sender->StatFlowRate(),
		sender->StatFlowLossRate());
}
