#include "Headers.h"
#include "HostServerPush.h"

CHostServerPush::CHostServerPush(
	__uint32		id,
	CSimHandler*	sim,
	CAddress		address,
	CInfoPush*		info,
	CData*			data
	) : CHostServer(id, sim, address, info, data)
{
	// Global
	this->info = info;

	// Delegates
	this->delegateRecv = new Delegate4<CHostServerPush, void, __uint32, CPacketIp*, CPacketUdp*, CPacket*>(this, &CHostServerPush::Recv);
	assert_mem(this->delegateRecv);

	this->delegateSendStream = new Delegate2<CHostServerPush, void, CAddress, CPacketStream*>(this, &CHostServerPush::SendStream);
	assert_mem(this->delegateSendStream);

	this->delegateSendMessage = new Delegate2<CHostServerPush, void, CAddress, CStreamMessage*>(this, &CHostServerPush::SendMessage);
	assert_mem(this->delegateSendMessage);

	// Create stream servers for multicast channels
	this->streamServersMcast = new CStreamServerPushMcast*[this->info->NumChannelsMulticast()];
	assert_mem(this->streamServersMcast);

	for(__uint32 index = 0; index < this->info->NumChannelsMulticast(); index++)
	{
		this->streamServersMcast[index] = new CStreamServerPushMcast(
			this->sim,
			this->info->StreamSource(this->info->IndexMulticastToGlobal(index)),
			this->delegateSendStream,
			this->layerIgmp->DelegateJoin(),
			this->layerIgmp->DelegateLeave(),
			0
			);
		assert_mem(this->streamServersMcast[index]);
	}

	// Create stream servers for unicast channels
	this->streamServersUcast = new CStreamServerPushUcast*[this->info->NumChannelsUnicast()];
	assert_mem(this->streamServersUcast);

	for(__uint32 index = 0; index < this->info->NumChannelsUnicast(); index++)
	{
		this->streamServersUcast[index] = new CStreamServerPushUcast(
			this->sim,
			this->info->StreamSource(this->info->IndexUnicastToGlobal(index)),
			this->delegateSendStream,
			this->delegateSendMessage
			);
		assert_mem(this->streamServersUcast[index]);
	}

	// Create layer connections
		// Layer UDP -> this
	(*this->layerUdp->EventRecv()) += this->delegateRecv;
}

CHostServerPush::~CHostServerPush()
{
	// Delete stream servers
	for(__uint32 index = 0; index < this->info->NumChannelsMulticast(); index++)
		delete this->streamServersMcast[index];
	delete[] this->streamServersMcast;

	for(__uint32 index = 0; index < this->info->NumChannelsUnicast(); index++)
		delete this->streamServersUcast[index];
	delete[] this->streamServersUcast;

	// Delegates
	delete this->delegateRecv;
	delete this->delegateSendStream;
	delete this->delegateSendMessage;
}

void CHostServerPush::Start()
{
	// Start the multicast channel servers
	for(__uint32 index = 0; index < this->info->NumChannelsMulticast(); index++)
		this->streamServersMcast[index]->Start();

	// Start the unicast channel servers
	for(__uint32 index = 0; index < this->info->NumChannelsUnicast(); index++)
		this->streamServersUcast[index]->Start();
}

void CHostServerPush::Stop()
{
	// Start the multicast channel servers
	for(__uint32 index = 0; index < this->info->NumChannelsMulticast(); index++)
		this->streamServersMcast[index]->Stop();

	// Start the unicast channel servers
	for(__uint32 index = 0; index < this->info->NumChannelsUnicast(); index++)
		this->streamServersUcast[index]->Stop();
}

void CHostServerPush::Finalize()
{
	// Call base class finalizer
	CHostServer::Finalize();

	// Call stream servers finalizer
	for(__uint32 index = 0; index < this->info->NumChannelsMulticast(); index++)
		this->streamServersMcast[index]->Finalize();
	for(__uint32 index = 0; index < this->info->NumChannelsUnicast(); index++)
		this->streamServersUcast[index]->Finalize();
}

void CHostServerPush::Recv(__uint32 entry, CPacketIp* ip, CPacketUdp* udp, CPacket* packet)
{
	// Receive UDP packets : process only control messages for unicast servers

	if(NULL == packet) return;
	if(udp->Dst() != this->info->PortControl()) return;
	if(packet->Type() != PACKET_TYPE_MESSAGE) return;

	// Convert the packet to a message
	CStreamMessage* message = type_cast<CStreamMessage*>(packet);

	// Send the message to the upper layer : the server for the specified stream
	this->streamServersUcast[this->info->IndexGlobalToUnicast(message->Stream())]->Recv(ip->Src(), message);
}

void CHostServerPush::SendStream(CAddress dst, CPacketStream* packet)
{
	// Send stream packet to the lower layer (TTL = 128)
	(*this->layerUdp->DelegateSend())(this->info->PortStream(), this->info->PortStream(), dst, 128, packet);
}

void CHostServerPush::SendMessage(CAddress dst, CStreamMessage* message)
{
	// Send message packet to the lower layer (TTL = 128)
	(*this->layerUdp->DelegateSend())(this->info->PortControl(), this->info->PortControl(), dst, 128, message);
}
