#include "Headers.h"
#include "LayerIgmp.h"

CLayerIgmp::CLayerIgmp(
	CSimHandler*	sim,
	CAddress		address,
	__uint32		numGroups,
	EIgmpType		type
	) : CLayer(sim)
{
	this->address = address;
	this->numGroups = numGroups;
	this->type = type;

	// Local groups
	this->memberships = new CLayerIpMcastMembership[this->numGroups];
	assert_mem(this->memberships);

	// Remote groups
	this->groups = new CLayerIpMcastGroup[this->numGroups];
	assert_mem(this->groups);

	// Delegates
	this->delegateRecv = new Delegate2<CLayerIgmp, void, __uint32, CPacketIp*>(this, &CLayerIgmp::Recv);
	assert_mem(this->delegateRecv);

	this->delegateJoin = new Delegate2<CLayerIgmp, void, CAddress, __uint32>(this, &CLayerIgmp::Join);
	assert_mem(this->delegateJoin);
	
	this->delegateLeave = new Delegate1<CLayerIgmp, void, CAddress>(this, &CLayerIgmp::Leave);
	assert_mem(this->delegateLeave);

	// Events
	this->eventJoin = new Event3<void, __uint32, CAddress, CLayerIpMcastGroup*>();
	assert_mem(this->eventJoin);

	this->eventLeave = new Event3<void, __uint32, CAddress, CLayerIpMcastGroup*>();
	assert_mem(this->eventLeave);

	this->eventLocalJoin = new Event2<void, __uint32, CAddress>();
	assert_mem(this->eventLocalJoin);

	this->eventLocalLeave = new Event2<void, __uint32, CAddress>();
	assert_mem(this->eventLocalLeave);

	this->eventSend = new Event2<void, __uint32, CPacketIp*>();
	assert_mem(this->eventSend);

	// Statistics
	this->statEntries = 0;
	this->statEntriesNum = 0;
	this->statEntriesLast = 0;
}

CLayerIgmp::~CLayerIgmp()
{
	delete[] this->memberships;
	delete[] this->groups;

	delete this->delegateRecv;
	delete this->delegateJoin;
	delete this->delegateLeave;

	delete this->eventJoin;
	delete this->eventLeave;
	
	delete this->eventLocalJoin;
	delete this->eventLocalLeave;

	delete this->eventSend;
}

void CLayerIgmp::Join(CAddress address, __uint32 entry)
{
	__uint32 group = address.MulticastGroup();
	assert(group < this->numGroups);

	// If the host is part of the multicast group, return
	if(this->memberships[group].State() == CLayerIpMcastMembership::MEMBER) return;

	// Send an unsolicited membership report to the IP multicast address of the group
	CPacketIgmp* packet = new CPacketIgmp(address, CPacketIgmp::PACKET_IGMP_MEMBERSHIP_REPORT);
	assert_mem(packet);

	// Send the IGMP packet on the joined interface, group multicast address and TTL of 1
	this->Send(entry, address, 1, packet);
	
	// Set the group state to joined
	this->memberships[group].Join(entry);

	// Call event for membership change
	(*this->eventLocalJoin)(entry, address);
}

void CLayerIgmp::Leave(CAddress address)
{
	__uint32 group = address.MulticastGroup();
	assert(group < this->numGroups);

	// If the host is not part of the multicast group, return
	if(this->memberships[group].State() == CLayerIpMcastMembership::NON_MEMBER) return;

	// Send a membership leave to the all routers multicast address
	CPacketIgmp* packet = new CPacketIgmp(address, CPacketIgmp::PACKET_IGMP_LEAVE_GROUP);
	assert_mem(packet);

	// Send the IGMP packet on the joined interface, all routers multicast address and TTL of 1
	this->Send(this->memberships[group].Entry(), ADDRESS_MCAST_ALL_ROUTERS, 1, packet);

	// Call event for membership change
	(*this->eventLocalLeave)(this->memberships[group].Entry(), address);

	// Set group state to leave
	this->memberships[group].Leave();
}

void CLayerIgmp::Recv(__uint32 entry, CPacketIp* packet)
{
	// Process only IGMP packet
	if(packet->Payload() == NULL) return;
	if(packet->Payload()->Type() != PACKET_TYPE_IGMP) return;

#ifdef LOG_LAYER
	printf("\n\tT = %7.3lf IGMP RECV local=%u bytes=%u src=%u dst=%u", this->sim->Time(), this->address, packet->Size(), packet->Src(), packet->Dst());
#endif

	switch(packet->Dst().Address())
	{
	case ADDRESS_MCAST_ALL_SYSTEMS: this->RecvIgmpAllSystems(entry, packet->Src(), packet->Dst(), type_cast<CPacketIgmp*>(packet->Payload())); break;
	case ADDRESS_MCAST_ALL_ROUTERS: this->RecvIgmpAllRouters(entry, packet->Src(), packet->Dst(), type_cast<CPacketIgmp*>(packet->Payload())); break;
	default: this->RecvIgmpOther(entry, packet->Src(), packet->Dst(), type_cast<CPacketIgmp*>(packet->Payload())); break;
	}
}

void CLayerIgmp::RecvIgmpAllSystems(__uint32 entry, CAddress src, CAddress dst, CPacketIgmp* packet)
{
	// Do nothing (current implementation does not process all systems IGMP packets)
}

void CLayerIgmp::RecvIgmpAllRouters(__uint32 entry, CAddress src, CAddress dst, CPacketIgmp* packet)
{
	// Process all routers IGMP packets if the IGMP type is router
	if(this->type != IGMP_ROUTER) return;

	switch(packet->IgmpType())
	{
	case CPacketIgmp::PACKET_IGMP_LEAVE_GROUP: this->RecvIgmpLeaveGroup(entry, src, dst, packet); break; // Leave group
	default:; // Do nothing
	}
}

void CLayerIgmp::RecvIgmpOther(__uint32 entry, CAddress src, CAddress dst, CPacketIgmp* packet)
{
	// Process only multicast addresses or unicast addressed to this host
	if((!dst.IsMulticast()) && (dst != this->address)) return;

	switch(packet->IgmpType())
	{
	case CPacketIgmp::PACKET_IGMP_MEMBERSHIP_REPORT: this->RecvIgmpJoinGroup(entry, src, dst, packet); break; // Join group
	default:; // Do nothing
	}
}

void CLayerIgmp::RecvIgmpJoinGroup(__uint32 entry, CAddress src, CAddress dst, CPacketIgmp* packet)
{
	// Join group (membership report)
	__uint32 group = dst.MulticastGroup();
	assert(group < this->numGroups);

	// Statistics
	this->statEntries += (this->sim->Time() - this->statEntriesLast) * this->statEntriesNum;
	this->statEntriesNum -= this->groups[group].Out()->size();

	// Add the interface to the group
	this->groups[group].Out()->insert(entry);

	// Statistics
	this->statEntriesNum += this->groups[group].Out()->size();
	this->statEntriesLast = this->sim->Time();

	// Call event for this group
	(*this->eventJoin)(entry, dst, &this->groups[group]);
}

void CLayerIgmp::RecvIgmpLeaveGroup(__uint32 entry, CAddress src, CAddress dst, CPacketIgmp* packet)
{
	// Leave group
	__uint32 group = packet->Group().MulticastGroup();
	assert(group < this->numGroups);

	// Statistics
	this->statEntries += (this->sim->Time() - this->statEntriesLast) * this->statEntriesNum;
	this->statEntriesNum -= this->groups[group].Out()->size();

	// Remove the interface from the group
	this->groups[group].Out()->erase(entry);

	// Statistics
	this->statEntriesNum += this->groups[group].Out()->size();
	this->statEntriesLast = this->sim->Time();

	// Call event for this group
	(*this->eventLeave)(entry, packet->Group(), &this->groups[group]);
}

void CLayerIgmp::Send(__uint32 entry, CAddress dst, __byte ttl, CPacket* payload)
{
	// Create a new IP packet
	CPacketIp* packet = new CPacketIp(this->address, dst, ttl, payload);
	assert_mem(packet);

#ifdef LOG_LAYER
	printf("\n\tT = %7.3lf IGMP SEND local=%u bytes=%u src=%u dst=%u", this->sim->Time(), this->address, packet->Size(), packet->Src(), packet->Dst());
#endif

	// Send the packet to the lower layer
	(*this->eventSend)(entry, packet);
}

void CLayerIgmp::Finalize()
{
	this->statEntries += (this->sim->Time() - this->statEntriesLast) * this->statEntriesNum;
	this->statEntries /= this->sim->Time();
}