#include "Headers.h"
#include "LayerIpMcast.h"

CLayerIpMcast::CLayerIpMcast(
	CSimHandler*	sim,
	__uint32		numGroups
	) : CLayer(sim)
{
	this->numGroups = numGroups;

	// Groups
	this->groups = new CLayerIpMcastGroup[this->numGroups];
	assert_mem(this->groups);

	// Memberships
	this->memberships = new CLayerIpMcastMembership[this->numGroups];
	assert_mem(this->groups);

	// Delegates
	this->delegateRecv = new Delegate2<CLayerIpMcast, void, __uint32, CPacketIp*>(this, &CLayerIpMcast::Recv);
	assert_mem(this->delegateRecv);

	this->delegateSend = new Delegate1<CLayerIpMcast, void, CPacketIp*>(this, &CLayerIpMcast::Send);
	assert_mem(this->delegateSend);

	this->delegateJoin =  new Delegate3<CLayerIpMcast, void, __uint32, CAddress, CLayerIpMcastGroup*>(this, &CLayerIpMcast::Join);
	assert_mem(this->delegateJoin);
	
	this->delegateLeave = new Delegate3<CLayerIpMcast, void, __uint32, CAddress, CLayerIpMcastGroup*>(this, &CLayerIpMcast::Leave);
	assert_mem(this->delegateLeave);

	this->delegateLocalJoin = new Delegate2<CLayerIpMcast, void, __uint32, CAddress>(this, &CLayerIpMcast::LocalJoin);
	assert_mem(this->delegateLocalJoin);
	
	this->delegateLocalLeave = new Delegate2<CLayerIpMcast, void, __uint32, CAddress>(this, &CLayerIpMcast::LocalLeave);
	assert_mem(this->delegateLocalLeave);

	// Events
	this->eventRecv = new Event2<void, __uint32, CPacketIp*>();
	assert_mem(this->eventRecv);

	this->eventSend = new Event2<void, __uint32, CPacketIp*>();
	assert_mem(this->eventSend);
}

CLayerIpMcast::~CLayerIpMcast()
{
	delete[] this->groups;
	delete[] this->memberships;

	// Delegates
	delete this->delegateRecv;
	delete this->delegateSend;
	delete this->delegateJoin;
	delete this->delegateLeave;
	delete this->delegateLocalJoin;
	delete this->delegateLocalLeave;

	// Event
	delete this->eventRecv;
	delete this->eventSend;
}

void CLayerIpMcast::Recv(__uint32 entry, CPacketIp* packet)
{
	// Process only multicast packets addressed to a multicast group
	if(!packet->Dst().IsMulticastGroup()) return;

	// Calculate the group
	__uint32 group = packet->Dst().MulticastGroup();
	assert(group < this->numGroups);

	/*
	 * Local forwarding
	 */
	// If there is a local membership for this group send the packet payload to the upper layers
	if(this->memberships[group].State() == CLayerIpMcastMembership::MEMBER)
	{
		(*this->eventRecv)(entry, packet);
	}

	/*
	 * Multicast routing
	 */
	// Execute multicast forwarding only if TTL is greater than zero
	if(packet->Ttl() == 0) return;

	// Forward a copy of the packet to all output interfaces for this group except the inbound interface
	for(CLayerIpMcastGroup::Iterator iter = this->groups[group].Out()->begin(); iter != this->groups[group].Out()->end(); iter++)
	{
		if(*iter != entry)
		{
			(*this->eventSend)(*iter, type_cast<CPacketIp*>(packet->Copy()));
		}
	}
}

void CLayerIpMcast::Send(CPacketIp* packet)
{
	// Process only packets addressed to a multicast group
	if(!packet->Dst().IsMulticastGroup()) return;

	// Get the group
	__uint32 group = packet->Dst().MulticastGroup();
	assert(group < this->numGroups);

	// Forward the packet only if there is a membership for this group
	if(this->memberships[group].State() != CLayerIpMcastMembership::MEMBER) return;

	// Send the packet to the lower layer on the membership interface
	(*this->eventSend)(this->memberships[group].Entry(), packet);
}

void CLayerIpMcast::Join(__uint32 entry, CAddress address, CLayerIpMcastGroup* groupSender)
{
	__uint32 group = address.MulticastGroup();
	assert(group < this->numGroups);

	// Add interface to multicast group
	this->groups[group].Out()->insert(entry);
}

void CLayerIpMcast::Leave(__uint32 entry, CAddress address, CLayerIpMcastGroup* groupSender)
{
	__uint32 group = address.MulticastGroup();
	assert(group < this->numGroups);

	// Remove interface from multicast group
	this->groups[group].Out()->erase(entry);
}

void CLayerIpMcast::LocalJoin(__uint32 entry, CAddress address)
{
	__uint32 group = address.MulticastGroup();
	assert(group < this->numGroups);

	// Add membership for this group and interface
	this->memberships[group].Join(entry);
}

void CLayerIpMcast::LocalLeave(__uint32 entry, CAddress address)
{
	__uint32 group = address.MulticastGroup();
	assert(group < this->numGroups);

	// Remove membership for this group and interface
	this->memberships[group].Leave();
}

