#include "Headers.h"
#include "LayerPimSm.h"

CLayerPimSm::CLayerPimSm(
	CSimHandler*	sim,
	__uint32		numGroups,
	CAddress		address,
	CAddress*		rp,
	CRoute*			route
	) : CLayer(sim)
{
	this->numGroups = numGroups;
	this->address = address;
	this->route = route;
	
	// Rendezvous-points
	this->rp = rp;

	// Memeberships (IGMP)
	this->memberships = new CLayerPimSmMembership[this->numGroups];
	assert_mem(this->memberships);

	// Groups (*, G)
	this->groups = new CLayerIpMcastGroup[this->numGroups];
	assert_mem(this->groups);

	// Delegates
	this->delegateRecv = new Delegate2<CLayerPimSm, void, __uint32, CPacketIp*>(this, &CLayerPimSm::Recv);
	assert_mem(this->delegateRecv);

	this->delegateJoin = new Delegate3<CLayerPimSm, void, __uint32, CAddress, CLayerIpMcastGroup*>(this, &CLayerPimSm::Join);
	assert_mem(this->delegateJoin);
	
	this->delegateLeave = new Delegate3<CLayerPimSm, void, __uint32, CAddress, CLayerIpMcastGroup*>(this, &CLayerPimSm::Leave);
	assert_mem(this->delegateLeave);

	// Events
	this->eventJoin = new Event3<void, __uint32, CAddress, CLayerIpMcastGroup*>();
	assert_mem(this->eventJoin);

	this->eventLeave = new Event3<void, __uint32, CAddress, CLayerIpMcastGroup*>();
	assert_mem(this->eventLeave);

	this->eventSend = new Event2<void, __uint32, CPacketIp*>();
	assert_mem(this->eventSend);

	// Statistics
	this->statEntries = 0;
	this->statEntriesNum = 0;
	this->statEntriesLast = 0;
}

CLayerPimSm::~CLayerPimSm()
{
	// Memberships (IGMP)
	delete[] this->memberships;

	// Groups (*, G)
	delete[] this->groups;

	// Delegates
	delete this->delegateRecv;
	delete this->delegateJoin;
	delete this->delegateLeave;

	// Events
	delete this->eventJoin;
	delete this->eventLeave;
	delete this->eventSend;
}

void CLayerPimSm::Recv(__uint32 entry, CPacketIp* packet)
{
	// Process only PIM-SM packets
	if(packet->Dst() != ADDRESS_MCAST_PIM_SM) return;
	if(packet->Payload() == NULL) return;
	if(packet->Payload()->Type() != PACKET_TYPE_PIM_SM) return;

#ifdef LOG_LAYER
	printf("\n\tT = %7.3lf PIM_SM RECV local=%u entry=%u bytes=%u src=%u dst=%u", this->sim->Time(), this->address, entry, packet->Size(), packet->Src(), packet->Dst());
#endif

	this->RecvPimSm(entry, packet->Src(), packet->Dst(), type_cast<CPacketPimSm*>(packet->Payload()));
}

void CLayerPimSm::RecvPimSm(__uint32 entry, CAddress src, CAddress dst, CPacketPimSm* packet)
{
	switch(packet->PimSmType())
	{
	case CPacketPimSm::JOIN_PRUNE: this->RecvPimSmJoinPrune(entry, src, dst, type_cast<CPacketPimSmJoinPrune*>(packet)); break;
	default:; // do nothing
	}
}

void CLayerPimSm::RecvPimSmJoinPrune(__uint32 entry, CAddress src, CAddress dst, CPacketPimSmJoinPrune* packet)
{
	// Get the group
	__uint32 group = packet->Address().MulticastGroup();
	assert(group < this->numGroups);

	switch(packet->JoinPruneType())
	{
	case CPacketPimSmJoinPrune::JOIN:
#ifdef LOG_LAYER
		printf("\n\tT = %7.3lf PIM_SM RECV local=%u JOIN %u", this->sim->Time(), this->address, group);
#endif

		// Join the group
		this->JoinGroup(group, packet->Address());

		// Statistics
		this->statEntries += (this->sim->Time() - this->statEntriesLast) * this->statEntriesNum;
		this->statEntriesNum -= this->groups[group].Out()->size();

		// Add the interface to the list of outbound interfaces for this group
		this->groups[group].Out()->insert(entry);

		// Statistics
		this->statEntriesNum += this->groups[group].Out()->size();
		this->statEntriesLast = this->sim->Time();

		// Call event for this group
		(*this->eventJoin)(entry, packet->Address(), &this->groups[group]);
		
		break;
	case CPacketPimSmJoinPrune::PRUNE:
#ifdef LOG_LAYER
		printf("\n\tT = %7.3lf PIM_SM RECV local=%u PRUNE %u", this->sim->Time(), this->address, group);
#endif

		// Statistics
		this->statEntries += (this->sim->Time() - this->statEntriesLast) * this->statEntriesNum;
		this->statEntriesNum -= this->groups[group].Out()->size();

		// Remove the interface from the list of outbound interfaces for this group
		this->groups[group].Out()->erase(entry);

		// Statistics
		this->statEntriesNum += this->groups[group].Out()->size();
		this->statEntriesLast = this->sim->Time();
	
		// Leave the group
		this->LeaveGroup(group, packet->Address());

		// Call event for this group
		(*this->eventLeave)(entry, packet->Address(), &this->groups[group]);

		break;
	}
}

void CLayerPimSm::Join(__uint32 entry, CAddress address, CLayerIpMcastGroup* igmp)
{
	__uint32 group = address.MulticastGroup();
	assert(group < this->numGroups);

	// Join the group
	this->JoinGroup(group, address);

	// Update group membership
	this->memberships[group] = igmp->Out()->size()?CLayerPimSmMembership::MEMBER:CLayerPimSmMembership::NON_MEMBER;
}

void CLayerPimSm::Leave(__uint32 entry, CAddress address, CLayerIpMcastGroup* igmp)
{
	__uint32 group = address.MulticastGroup();
	assert(group < this->numGroups);

	// Update group membership
	this->memberships[group] = igmp->Out()->size()?CLayerPimSmMembership::MEMBER:CLayerPimSmMembership::NON_MEMBER;

	// Leave the group
	this->LeaveGroup(group, address);
}

void CLayerPimSm::JoinGroup(__uint32 group, CAddress address)
{
	// If the group is new (i.e. no output interfaces) and this is not the RP router for this group and there is no (IGMP) membership
	if((this->groups[group].Out()->size() == 0) && (this->rp[group] != this->address) && (this->memberships[group].State() == CLayerPimSmMembership::NON_MEMBER))
	{
		// Send a JOIN message toward the RP router
		CPacketPimSmJoinPrune* join = new CPacketPimSmJoinPrune(
			CPacketPimSmJoinPrune::JOIN,
			address
			);
		assert_mem(join);

		// Send the packet on the link toward the RP with TTL of 1
		__uint32 link = this->route->Forward(this->address, this->rp[group]);

		this->Send(link, ADDRESS_MCAST_PIM_SM, 1, join);
	}
}

void CLayerPimSm::LeaveGroup(__uint32 group, CAddress address)
{
	// If the group is empty (i.e. no output interfaces) and this is not the RP router for this group and there are no memberships (IGMP)
	if((this->groups[group].Out()->size() == 0) && (this->rp[group] != this->address) && (this->memberships[group].State() == CLayerPimSmMembership::NON_MEMBER))
	{
		// Send a PRUNE message toward the RP router
		CPacketPimSmJoinPrune* prune = new CPacketPimSmJoinPrune(
			CPacketPimSmJoinPrune::PRUNE,
			address
			);
		assert_mem(prune);

		// Send the packet on the link toward the RP with TTL of 1
		__uint32 link = this->route->Forward(this->address, this->rp[group]);

		this->Send(link, ADDRESS_MCAST_PIM_SM, 1, prune);
	}
}

void CLayerPimSm::Send(__uint32 entry, CAddress dst, __byte ttl, CPacket* payload)
{
	// Create a new IP packet
	CPacketIp* packet = new CPacketIp(this->address, dst, ttl, payload);
	assert_mem(packet);

#ifdef LOG_LAYER
	printf("\n\tT = %7.3lf PIM_SM SEND local=%u entry=%u bytes=%u src=%u dst=%u", this->sim->Time(), this->address, entry, packet->Size(), packet->Src(), packet->Dst());
#endif

	// Send the packet to the lower layer
	(*this->eventSend)(entry, packet);
}

void CLayerPimSm::Finalize()
{
	this->statEntries += (this->sim->Time() - this->statEntriesLast) * this->statEntriesNum;
	this->statEntries /= this->sim->Time();
}