#include "Headers.h"
#include "Link.h"
#include "EventLinkQueue.h"
#include "EventLinkDelay.h"
#include "Console.h"

#define LINK_NODE_OUT(in) (in ^ 1)

CLink::CLink(
	__uint32		id,
	CSimHandler*	sim,
	__uint32		queue,
	__bitrate		bandwidth0,
	__bitrate		bandwidth1,
	__time			delay0,
	__time			delay1
	) : CObject(sim)
{
	this->id = id;

	assert(bandwidth);
	this->bandwidth[0] = bandwidth0;
	this->bandwidth[1] = bandwidth1;
	this->delay[0] = delay0;
	this->delay[1] = delay1;

	this->nodes[0] = NULL;
	this->nodes[1] = NULL;

	this->nodeEntries[0] = 0;
	this->nodeEntries[1] = 0;

	// Low priority queues
	this->queues[0][0] = new CLinkQueue(queue);
	assert_mem(this->queues[0][0]);
	this->queues[0][1] = new CLinkQueue(queue);
	assert_mem(this->queues[0][1]);
	// High priority queues
	this->queues[1][0] = new CLinkQueue(queue);
	assert_mem(this->queues[1][0]);
	this->queues[1][1] = new CLinkQueue(queue);
	assert_mem(this->queues[1][1]);

	this->currentNode = 0;

	// Utilization meter
	this->meterTime[0] = 0;
	this->meterTime[1] = 0;
	this->meterUtil[0] = 0;
	this->meterUtil[1] = 0;

	// Statistics
	this->statUtil[0][0] = 0;
	this->statUtil[0][1] = 0;
	this->statUtil[1][0] = 0;
	this->statUtil[1][1] = 0;
	
	this->statQueue[0][0] = 0;
	this->statQueue[0][1] = 0;
	this->statQueue[1][0] = 0;
	this->statQueue[1][1] = 0;

	this->statQueueLast[0][0] = 0;
	this->statQueueLast[0][1] = 0;
	this->statQueueLast[1][0] = 0;
	this->statQueueLast[1][1] = 0;

	this->statPackets[0][0] = 0;
	this->statPackets[0][1] = 0;
	this->statPackets[1][0] = 0;
	this->statPackets[1][1] = 0;

	this->statBits[0][0] = 0;
	this->statBits[0][1] = 0;
	this->statBits[1][0] = 0;
	this->statBits[1][1] = 0;

	this->statDiscard[0][0] = 0;
	this->statDiscard[0][1] = 0;
	this->statDiscard[1][0] = 0;
	this->statDiscard[1][1] = 0;
}

CLink::~CLink()
{
	delete this->queues[0][0];
	delete this->queues[0][1];
	delete this->queues[1][0];
	delete this->queues[1][1];
}

double CLink::meterSmoothFactor = 0.9;

__uint32 CLink::AddNode(CObject* node, __uint32 entry)
{
	assert(node);
	assert(this->currentNode < 2);

	this->nodes[this->currentNode] = node;
	this->nodeEntries[this->currentNode] = entry;

	this->currentNode++;
	
	return this->currentNode-1;
}

ENetworkCode CLink::Recv(CObject* sender, __uint32 entry, CPacket* packet)
{
	assert(sender);
	assert(packet);
	assert((0 == entry) || (1 == entry));
	assert(this->nodes[entry] == sender);

	__byte tos = packet->Tos();

#ifdef LOG_LINK
	CConsole::SetColor(CConsole::LIGHT_YELLOW);
	printf("\n%10.6lf : LINK_%x RECV entry=%u size=%u tos=%u queue=%u/%u", this->sim->Time(), this->id, entry, packet->Size(), tos, this->queues[tos][entry]->Packets(), this->queues[tos][entry]->Size());
	CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif

	// Stat : packets
	this->statPackets[tos][entry]++;

	// If the queue is full discard the message
	if(this->queues[tos][entry]->IsFull())
	{
#ifdef LOG_LINK
		CConsole::SetColor(CConsole::LIGHT_YELLOW);
		printf("\n%10.6lf : LINK_%x DISCARD tos=%u", this->sim->Time(), this->id, tos);
		CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif

		// Verify that high priority packets are not discarded
		assert(tos == CPacket::PACKET_TOS_LOW);

		// Stat : discarded packets
		this->statDiscard[tos][entry]++;

		packet->Delete();
		delete packet;
		return NETWORK_FAIL;
	}

	// If all queues are empty, schedule a link event now to transmit the message
	if(this->queues[0][entry]->IsEmpty() && this->queues[1][entry]->IsEmpty())
	{
#ifdef LOG_LINK
		CConsole::SetColor(CConsole::LIGHT_YELLOW);
		printf("\n%10.6lf : LINK_%x START SEND size=%u tos=%u bw=%lf delay=%lf (forw=%lf tran=%lf)", this->sim->Time(), this->id, packet->Size(), tos, this->bandwidth[entry], this->delay[entry] + (packet->Size() / this->bandwidth[entry]), this->delay[entry], (packet->Size() / this->bandwidth[entry]));
		CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif
		__time delay = (packet->Size() + 1) / this->bandwidth[entry];

		CEventLinkQueue* evt = new CEventLinkQueue(this, entry, packet);
		assert_mem(evt);

		this->sim->ScheduleEventAfter(
			delay,
			evt
			);

		// Stat : link utlization
		this->statUtil[tos][entry] += packet->Size() / this->bandwidth[entry];

		// Stat : data
		this->statBits[tos][entry] += packet->Size();

		// Utilization meter
		this->meterUtil[entry] = CLink::meterSmoothFactor * packet->Size() / (this->sim->Time() + delay - this->meterTime[entry]) +
			(1 - CLink::meterSmoothFactor) * this->meterUtil[entry];
		this->meterTime[entry] = this->sim->Time() + delay;
	}
#ifdef LOG_LINK
	else
	{
		CConsole::SetColor(CConsole::LIGHT_YELLOW);
		printf("\n%10.6lf : LINK_%x ENQUEUE", this->sim->Time(), this->id);
		CConsole::SetColor(CConsole::LIGHT_GRAY);
	}
#endif

	// Stat : queue
	this->statQueue[tos][entry] += this->queues[tos][entry]->Packets() * (this->sim->Time() - this->statQueueLast[tos][entry]);
	this->statQueueLast[tos][entry] = this->sim->Time();

	// Add the message to the queue
	this->queues[tos][entry]->Add(this->sim->Time(), packet);

	return NETWORK_SUCCESS;
}

void CLink::Delay(__uint32 entry, CPacket* packet)
{
	assert(packet);

	__byte tos = packet->Tos();

	assert(this->queues[tos][entry]->First() == packet);

	// Stat : queue
	this->statQueue[tos][entry] += this->queues[tos][entry]->Packets() * (this->sim->Time() - this->statQueueLast[tos][entry]);
	this->statQueueLast[tos][entry] = this->sim->Time();

	// Remove packet from the queue
	this->queues[tos][entry]->Remove();

#ifdef LOG_LINK
	CConsole::SetColor(CConsole::LIGHT_YELLOW);
	printf("\n%10.6lf : LINK_%x DELAY SEND size=%u tos=%u", this->sim->Time(), this->id, packet->Size(), tos);
	CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif

	// Delay the packet
	CEventLinkDelay* evt = new CEventLinkDelay(this, entry, packet);
	assert_mem(evt);

	this->sim->ScheduleEventAfter(
		this->delay[entry],
		evt);

	// If the queue is not empty, process the next packet starting with the high priority queue
	for(__byte tos = CPacket::PACKET_TOS_HIGH; tos <= CPacket::PACKET_TOS_LOW; tos++)
	{
		if(!this->queues[tos][entry]->IsEmpty())
		{
			assert(this->queues[tos][entry]->First());

			CPacket* packet = this->queues[tos][entry]->First();

			assert(packet->Tos() == tos);

#ifdef LOG_LINK
			CConsole::SetColor(CConsole::LIGHT_YELLOW);
			printf("\n%10.6lf : LINK_%x START SEND size=%u tos=%u bw=%lf delay=%lf (forw=%lf tran=%lf)", this->sim->Time(), this->id, packet->Size(), tos, this->bandwidth[entry], this->delay[entry] + (packet->Size() / this->bandwidth[entry]), this->delay[entry], (packet->Size() / this->bandwidth[entry]));
			CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif
			__time delay = (packet->Size() + 1) / this->bandwidth[entry];


			CEventLinkQueue* evt = new CEventLinkQueue(this, entry, packet);
			assert_mem(evt);

			this->sim->ScheduleEventAfter(
				delay,
				evt
				);

			// Stat : link utlization
			this->statUtil[tos][entry] += packet->Size() / this->bandwidth[entry];

			// Stat : data
			this->statBits[tos][entry] += packet->Size();

			// Utilization meter
			this->meterUtil[entry] = CLink::meterSmoothFactor * packet->Size() / (this->sim->Time() + delay - this->meterTime[entry]) +
			(1 - CLink::meterSmoothFactor) * this->meterUtil[entry];
			this->meterTime[entry] = this->sim->Time() + delay;
	
			// If one packet is transmitted, stop the loop
			break;
		}
	}
}

void CLink::Send(__uint32 entry, CPacket* packet)
{
	__uint32 out = LINK_NODE_OUT(entry);

#ifdef LOG_LINK
	CConsole::SetColor(CConsole::LIGHT_YELLOW);
	printf("\n%10.6lf : LINK_%x FINISH SEND entry=%u size=%u tos=%u", this->sim->Time(), this->id, out, packet->Size(), packet->Tos());
	CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif

	// Call receive function of the destination node
	this->nodes[out]->Recv(this, this->nodeEntries[out], packet);
}

__bitrate CLink::MeterUtil(__uint8 entry)
{
	return this->meterUtil[entry];
}

void CLink::Finalize()
{
	this->statUtil[0][0] /= this->sim->Time();
	this->statUtil[0][1] /= this->sim->Time();
	this->statUtil[1][0] /= this->sim->Time();
	this->statUtil[1][1] /= this->sim->Time();

	this->statQueue[0][0] += this->queues[0][0]->Packets() * (this->sim->Time() - this->statQueueLast[0][0]);
	this->statQueue[0][1] += this->queues[0][1]->Packets() * (this->sim->Time() - this->statQueueLast[0][1]);
	this->statQueue[1][0] += this->queues[1][0]->Packets() * (this->sim->Time() - this->statQueueLast[1][0]);
	this->statQueue[1][1] += this->queues[1][1]->Packets() * (this->sim->Time() - this->statQueueLast[1][1]);

	this->statQueue[0][0] /= this->sim->Time();
	this->statQueue[0][1] /= this->sim->Time();
	this->statQueue[1][0] /= this->sim->Time();
	this->statQueue[1][1] /= this->sim->Time();
}
