#include "Headers.h"
#include "MulticastTree.h"
#include "Console.h"
#include "Debug.h"

CMulticastTree::CMulticastTree(CTopo* topology, __uint32 group, __uint32 source)
{
	assert(topology);

	this->topology = topology;
	this->source = source;
	this->group = group;

	// Create tree nodes
	this->nodes = new CMulticastNode[this->topology->Nodes()];
	assert_mem(this->nodes);

	// Stat
	this->sizeCoreLast = 0;
	this->sizeCoreTime = 0;
	this->sizeCore = 0;

	this->sizeEdgeLast = 0;
	this->sizeEdgeTime = 0;
	this->sizeEdge = 0;
}

CMulticastTree::~CMulticastTree()
{
	delete[] this->nodes;
}

void CMulticastTree::Add(__time time, __uint32 host, __uint32 gw)
{
	assert(gw < this->topology->Nodes());

	// Check the host is not attached to the gateway node
	assert(this->nodes[gw].Hosts()->find(host) == this->nodes[gw].Hosts()->end());

#ifdef LOG_MCAST
	if(LOG_GROUP == this->group) printf("\nT = %7.3lf GROUP %u ADD [%u] --- (%u)", time, this->group, host, gw);
#endif

	// Add the host to the node
	this->nodes[gw].Hosts()->insert(host);

	// Calculate stat
	this->sizeEdge += (time - this->sizeEdgeTime) * this->sizeEdgeLast;
	this->sizeEdgeTime = time;

	this->sizeCore += (time - this->sizeCoreTime) * this->sizeCoreLast;
	this->sizeCoreTime = time;

	// Update edge stat
	this->sizeEdgeLast++;

	// Generate the path from the host node to the source
	__uint32 node = gw;
	__uint32 next = node;
	while(node != this->source)
	{
		// Get the next node toward the destination
		next = this->topology->Route()->NextNode(node, this->source);

		// Check if next node is already part of the group
		bool member = (this->nodes[next].Out()->size() > 0) || (this->nodes[next].Hosts()->size() > 0);

		// If the next node has the current node in the output list, stop
		if(this->nodes[next].Out()->find(node) != this->nodes[next].Out()->end())
		{
#ifdef LOG_MCAST
			CConsole::SetColor(CConsole::LIGHT_CYAN);
			if(LOG_GROUP == this->group) printf(" --> (%u)", next);
			CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif
			break;
		}
#ifdef LOG_MCAST
		else
		{
			CConsole::SetColor(CConsole::LIGHT_GREEN);
			if(LOG_GROUP == this->group) printf(" --> (%u)", next);
			CConsole::SetColor(CConsole::LIGHT_GRAY);
		}
#endif

		// Else, add the current node to the next node
		this->nodes[next].Out()->insert(node);

		// Update core stat
		this->sizeCoreLast++;

		// If the next node was already a member, stop
		if(member) break;

		// Move to next node
		node = next;
	}

#ifdef LOG_MCAST
	// If current node is not the source, display path to the source
	for(node = next; node != this->source; )
	{
		next = this->topology->Route()->NextNode(node, this->source);

		CConsole::SetColor(CConsole::LIGHT_CYAN);
		if(LOG_GROUP == this->group) printf(" --> (%u)", next);
		CConsole::SetColor(CConsole::LIGHT_GRAY);

		node = next;
	}

	if(LOG_GROUP == this->group) printf("\n%3u\t\t%3u\t\t%7.3lf\t\t%7.3lf", this->sizeCoreLast, this->sizeEdgeLast, this->sizeCore/time, this->sizeEdge/time);
#endif
}

void CMulticastTree::Remove(__time time, __uint32 host, __uint32 gw)
{
#ifdef LOG_MCAST
	if(LOG_GROUP == this->group) printf("\nT = %7.3lf GROUP %u REM [%u]", time, this->group, host);
	CConsole::SetColor(CConsole::LIGHT_RED);
	if(LOG_GROUP == this->group) printf(" -x- (%u)", gw);
	CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif

	// Check the host is attached to the gateway node
	assert(this->nodes[gw].Hosts()->find(host) != this->nodes[gw].Hosts()->end());

	// Remove the host from the node
	this->nodes[gw].Hosts()->erase(host);

	// Calculate stat
	this->sizeEdge += (time - this->sizeEdgeTime) * this->sizeEdgeLast;
	this->sizeEdgeTime = time;

	this->sizeCore += (time - this->sizeCoreTime) * this->sizeCoreLast;
	this->sizeCoreTime = time;

	// Update edge stat
	this->sizeEdgeLast--;

	// If the hosts list is not empty, stop
	if(this->nodes[gw].Hosts()->size())
	{
#ifdef LOG_MCAST
		if(LOG_GROUP == this->group) printf("\n%3u\t\t%3u\t\t%7.3lf\t\t%7.3lf", this->sizeCoreLast, this->sizeEdgeLast, this->sizeCore/time, this->sizeEdge/time);
#endif
		return;
	}

	// If the out list is not empty, stop
	if(this->nodes[gw].Out()->size())
	{
#ifdef LOG_MCAST
		if(LOG_GROUP == this->group) printf("\n%3u\t\t%3u\t\t%7.3lf\t\t%7.3lf", this->sizeCoreLast, this->sizeEdgeLast, this->sizeCore/time, this->sizeEdge/time);
#endif
		return;
	}

	// Remove nodes from the tree
	for(__uint32 node = gw; node != this->source;)
	{
		// Get the next node
		__uint32 next = this->topology->Route()->NextNode(node, this->source);

		// Check the next node has the current node
		assert(this->nodes[next].Out()->find(node) != this->nodes[next].Out()->end());

		// Remove the current node from the next node
		this->nodes[next].Out()->erase(node);

		// Update core stat
		this->sizeCoreLast--;

#ifdef LOG_MCAST
		CConsole::SetColor(CConsole::LIGHT_RED);
		if(LOG_GROUP == this->group) printf(" -x- (%u)", next);
		CConsole::SetColor(CConsole::LIGHT_GRAY);
#endif

		// If the next node has other hosts, stop
		if(this->nodes[next].Hosts()->size()) break;

		// If the next node has other nodes, stop
		if(this->nodes[next].Out()->size()) break;

		// Else, go to next node
		node = next;
	}

#ifdef LOG_MCAST
	if(LOG_GROUP == this->group) printf("\n%3u\t\t%3u\t\t%7.3lf\t\t%7.3lf", this->sizeCoreLast, this->sizeEdgeLast, this->sizeCore/time, this->sizeEdge/time);
#endif
}

void CMulticastTree::Clear()
{
	for(__uint32 index = 0; index < this->topology->Nodes(); index++)
	{
		this->nodes[index].Hosts()->clear();
		this->nodes[index].Out()->clear();
	}

	// Stat
	this->sizeCoreLast = 0;
	this->sizeCoreTime = 0;
	this->sizeCore = 0;

	this->sizeEdgeLast = 0;
	this->sizeEdgeTime = 0;
	this->sizeEdge = 0;
}

void CMulticastTree::Finalize(__time time)
{
	this->sizeCore += (time - this->sizeCoreTime) * this->sizeCoreLast;
	this->sizeEdge += (time - this->sizeEdgeTime) * this->sizeEdgeLast;

	this->sizeCore /= time;
	this->sizeEdge /= time;
}