#include "Headers.h"
#include "Node.h"

CNode::CNode(
	__uint32		id,
	CSimHandler*	sim,
	CAddress		address,
	__uint32		numLinks,
	CRoute*			route,
	CInfo*			info
	) : CObject(sim)
{
	// Global
	this->id = id;
	this->address = address;
	this->info = info;
	this->route = route;

	// Links
	this->numLinks = numLinks;
	this->currentLink = 0;

	// Links
	this->links = new CLink*[this->numLinks];
	assert_mem(this->links);
	this->linkEntries = new __uint32[this->numLinks];
	assert_mem(this->linkEntries);

	// Base layer delegate and event
	this->delegateSend = new Delegate2<CNode, void, __uint32, CPacket*>(this, &CNode::Send);
	assert_mem(this->delegateSend);

	this->eventRecv = new Event2<void, __uint32, CPacket*>();
	assert_mem(this->eventRecv);

	// Initialize layers
	this->layerIp = new CLayerIp(this->sim);
	assert_mem(this->layerIp);

	this->layerIpLocal = new CLayerIpLocal(this->sim, this->address, this->route);
	assert_mem(this->layerIpLocal);

	this->layerIpMcast = new CLayerIpMcast(this->sim, this->info->NumGroups());
	assert_mem(this->layerIpMcast);

	this->layerUdp = new CLayerUdp(this->sim);
	assert_mem(this->layerUdp);

	// Create layer connections
		// Base layer <-> IP layer
	(*this->eventRecv) += this->layerIp->DelegateRecv();
	(*this->layerIp->EventSend()) += this->delegateSend;
		// IP layer <-> IP local layer
	(*this->layerIp->EventRecv()) += this->layerIpLocal->DelegateRecv();
	(*this->layerIpLocal->EventSend()) += this->layerIp->DelegateSend();
		// IP local layer <-> UDP layer
	(*this->layerIpLocal->EventRecv()) += this->layerUdp->DelegateRecv();
	(*this->layerUdp->EventSend()) += this->layerIpLocal->DelegateSend();
		// IP layer <-> IP multicast layer
	(*this->layerIp->EventRecv()) += this->layerIpMcast->DelegateRecv();
	(*this->layerIpMcast->EventSend()) += this->layerIp->DelegateSend();
		// IP local layer <-> IP multicast layer
	(*this->layerIpMcast->EventRecv()) += this->layerIpLocal->DelegateRecvMcast();
	(*this->layerIpLocal->EventSendMcast()) += this->layerIpMcast->DelegateSend();

	// Statistics
	this->statPacketsRead = 0;
	this->statPacketsWrite = 0;

	this->statDataRead = 0;
	this->statDataWrite = 0;
}

CNode::~CNode()
{
	// Delete base delegate and event
	delete this->delegateSend;
	delete this->eventRecv;

	// Delete layers
	delete this->layerIp;
	delete this->layerIpLocal;
	delete this->layerIpMcast;
	delete this->layerUdp;

	// Delete links
	delete[] this->links;
	delete[] this->linkEntries;
}

void CNode::AddLink(CLink* link)
{
	assert(link);
	assert(this->currentLink < this->numLinks);

	// Add link to the node
	this->links[this->currentLink] = link;
	
	// Add node to the link
	this->linkEntries[this->currentLink] = link->AddNode(this, this->currentLink);

	// Add link index
	this->linkIndices.insert(pair<__uint32, __uint32>(link->Id(), this->currentLink));

	this->currentLink++;
}

ENetworkCode CNode::Recv(CObject* sender, __uint32 entry, CPacket* packet)
{
	assert(sender);
	assert(packet);
	assert(sender == this->links[entry]);

#ifdef LOG_LAYER
	printf("\nT = %7.3lf NODE RECV id=%u entry=%u bytes=%u", this->sim->Time(), this->id, entry, packet->Size());
#endif

	// Send the packet to the upper layers
	(*this->eventRecv)(entry, packet);

	// Stat
	this->statPacketsRead++;
	this->statDataRead += packet->Size();

	// Delete the packet
	packet->Delete();
	delete packet;

	return NETWORK_SUCCESS;
}

void CNode::Send(__uint32 entry, CPacket* packet)
{
	assert(packet);
	assert(entry < this->numLinks);

	this->statPacketsWrite++;
	this->statDataWrite += packet->Size();

#ifdef LOG_LAYER
	printf("\nT = %7.3lf NODE SEND id=%u entry=%u bytes=%u", this->sim->Time(), this->id, entry, packet->Size());
#endif

	this->links[entry]->Recv(this, this->linkEntries[entry], packet);
}

void CNode::Finalize()
{
	// Call the finalizer for all layers
	this->layerIp->Finalize();
	this->layerIpLocal->Finalize();
	this->layerIpMcast->Finalize();
	this->layerUdp->Finalize();
}