#pragma once

#include "StreamClient.h"
#include "StreamBuffer.h"
#include "StreamEncoderFrame.h"
#include "StreamDecoderFrame.h"
#include "InfoPush.h"
#include "Link.h"
#include "Timer.h"

#include "StreamMessagePushJoin.h"
#include "StreamMessagePushLeave.h"
#include "StreamMessagePushClose.h"
#include "StreamMessageBootPushRequest.h"
#include "StreamMessageBootPushResponse.h"
#include "StreamMessageBootPushRegister.h"
#include "StreamMessageBootPushDeregister.h"

class CStreamClientPush : public CStreamClient
{
private:
	enum EClientState
	{
		STOPPED = 0,
		MCAST_STREAM_FIRST = 10,
		MCAST_STREAM_BUFFERING = 11,
		MCAST_STREAM_PLAY = 12,
		MCAST_STREAM_WAIT = 13,
		UCAST_BOOTSTRAP_REQUEST = 20,
		UCAST_BOOTSTRAP_RESPONSE = 21,
		UCAST_JOIN_REQUEST = 22,
		UCAST_STREAM_FIRST = 23,
		UCAST_STREAM_BUFFERING = 24,
		UCAST_STREAM_PLAY = 25,
		UCAST_STREAM_WAIT = 26,
		UCAST_STREAM_RESET = 27
	};

	enum ERegistrationState
	{
		NOT_REGISTERED = 0,
		REGISTERED = 1
	};

	// State
	EClientState					stateClient;
	ERegistrationState				stateRegistration;

	// Simulator
	CInfoPush*						info;
	CAddress						address;

	// Channel
	CChannel*						channel;

	// Buffer
	CStreamBuffer*					buffer;

	// Session
	void							(CStreamClientPush::*processFrame)(CStreamFrame frame);
	__time							sessionTimeLastStart;
	__time							sessionTimeLastStop;
	__uint32						sessionFrameLastStart;
	__uint32						sessionFrameLastStop;

	// Unicast session
	__bitrate						bw;				// Uplink bandwidth
	__bitrate						bwUsed;			// Used uplink bandwidth

	set<CAddress>					neighbors;		// Client neighbor list
	set<CAddress>					receivers;		// Client receiver list
	CAddress						sender;			// Client sender

	// Multicast session
	__uint32						entry;			// Multicast interface (0 - as host have only one interface)

	// Timer
	CTimer<CStreamClientPush>*		timerPlayMcast;	// Playback timer (multicast channels)
	CTimer<CStreamClientPush>*		timerPlayUcast; // Playback timer (unicast channels)
	CTimer<CStreamClientPush>*		timerBoot;		// Bootstrap timer
	CTimer<CStreamClientPush>*		timerRegister;	// Registrap timer
	CTimer<CStreamClientPush>*		timerBuffer;	// Buffering timer

	__uint32						timerBufferChannel;

	// Delegates - receive 
	Delegate2<CStreamClientPush, void, CAddress, CStreamFrame>*
									delegateRecvFrameMcast;
	Delegate2<CStreamClientPush, void, CAddress, CStreamFrame>*
									delegateRecvFrameUcast;

	IDelegate2<void, CAddress, CPacketStream*>*
									delegateSendStream;
	IDelegate2<void, CAddress, CStreamMessage*>*
									delegateSendMessage;

	IDelegate2<void, CAddress, __uint32>*
									delegateIgmpJoin;
	IDelegate1<void, CAddress>*		delegateIgmpLeave;


	IDelegate0<CLink*>*				delegateLink;


	// Encoder
	CStreamEncoderFrame*				encoder;

	// Decoders
	CStreamDecoder*					decoder;
	CStreamDecoderFrame*				decoderMcast;
	CStreamDecoderFrame*				decoderUcast;

	// Statistics
	__uint32						statRecvFrames;
	__uint32						statDiscardedFrames;
	__uint32						statPlayFrames;
	__uint32						statSuccessFrames[3];
	__uint32						statFailFrames[3];
	
	__uint32						statPlayFirstFrame;
	__uint32						statPlayLastFrame;

	__time							statTimeClientStart;
	__time							statTimeRecvStart;
	__time							statTimePlayStart;
	__time							statTimeFinish;
	__time							statTimeWait;

	__time							statSyncDelaySum;
	__uint32						statSyncDelayCount;

public:
	CStreamClientPush(
		CSimHandler*									sim,
		CInfoPush*										info,
		CAddress										address,
		IDelegate2<void, CAddress, CPacketStream*>*		delegateSendStream,
		IDelegate2<void, CAddress, CStreamMessage*>*	delegateSendMessage,
		IDelegate2<void, CAddress, __uint32>*			delegateIgmpJoin,
		IDelegate1<void, CAddress>*						delegateIgmpLeave,
		IDelegate0<CLink*>*								delegateLink,
		__uint32										entry,
		__bitrate										bw
		);
	virtual ~CStreamClientPush();

	virtual void				Start(CChannel* channel);
	virtual void				Stop();

	virtual void				Recv(CAddress srcAddress, CAddress dstAddress, __uint16 srcPort, __uint16 dstPort, CPacket* packet);
	
	virtual inline __uint32		StatRecvFrames() { return this->statRecvFrames; }
	virtual inline __uint32		StatDiscardedFrames() { return this->statDiscardedFrames; }
	virtual inline __uint32		StatPlayFrames() { return this->statPlayFrames; }
	virtual inline __uint32*	StatSuccessFrames() { return this->statSuccessFrames; }
	virtual inline __uint32*	StatFailFrames() { return this->statFailFrames; }
	
	virtual inline __time		StatTimeClientStart() { return this->statTimeClientStart; }
	virtual inline __time		StatTimeRecvStart() { return this->statTimeRecvStart; }
	virtual inline __time		StatTimePlayStart() { return this->statTimePlayStart; }
	virtual inline __time		StatTimeFinish() { return this->statTimeFinish; }
	virtual inline __time		StatTimeWait() { return this->statTimeWait; }

	virtual inline __uint32		StatPlayFirstFrame() { return this->statPlayFirstFrame; }
	virtual inline __uint32		StatPlayLastFrame() { return this->statPlayLastFrame; }

	virtual inline __time		StatSyncDelay() { return this->statSyncDelayCount?(this->statSyncDelaySum / this->statSyncDelayCount):-1; }

	virtual void				Finalize();

private:
	void						StartMcast();
	void						StartUcast();

	void						StopMcast();
	void						StopUcast();

	void						RecvStream(CAddress dst, CPacketStream* packet);
	void						RecvMessage(CAddress src, CStreamMessage* message);
	
	void						RecvFrameMcast(CAddress src, CStreamFrame frame);

	void						ProcessFrameMcastFirst(CStreamFrame frame);
	void						ProcessFrameMcastBuffering(CStreamFrame frame);
	void						ProcessFrameMcastPlay(CStreamFrame frame);
	void						ProcessFrameMcastWait(CStreamFrame frame);

	void						RecvFrameUcast(CAddress src, CStreamFrame frame);

	void						ProcessFrameUcastFirst(CStreamFrame frame);	
	void						ProcessFrameUcastBuffering(CStreamFrame frame);
	void						ProcessFrameUcastPlay(CStreamFrame frame);
	void						ProcessFrameUcastWait(CStreamFrame frame);
	void						ProcessFrameUcastReset(CStreamFrame frame);

	void						RecvMessageJoin(CAddress src, CStreamMessagePushJoin* message);
	void						RecvMessageLeave(CAddress src, CStreamMessagePushLeave* message);
	void						RecvMessageClose(CAddress src, CStreamMessagePushClose* message);
	void						RecvMessageBootResponse(CAddress src, CStreamMessageBootPushResponse* message);

	void						StreamBootstrap();
	void						StreamJoin();
	void						StreamLeave();
	void						StreamRegister();
	void						StreamDeregister();
	void						StreamClose();
	
	void						SendStream(CStreamFrame frame);

	void						RejectJoin(CAddress src, CStreamMessagePushJoin* message);

	void						TimerPlayMcast(CTimerInfo* info);
	void						TimerPlayUcast(CTimerInfo* info);
	void						TimerBoot(CTimerInfo* info);
	void						TimerRegister(CTimerInfo* info);
	void						TimerBuffer(CTimerInfo* info);
};