#pragma once

#include "StreamClient.h"
#include "StreamBuffer.h"
#include "StreamBufferMulti.h"
#include "StreamEncoderFrame.h"
#include "StreamDecoderFrame.h"
#include "InfoPushMulti.h"
#include "Link.h"
#include "Timer.h"

#include "StreamMessagePushMultiJoin.h"
#include "StreamMessagePushMultiLeave.h"
#include "StreamMessagePushMultiClose.h"
#include "StreamMessageBootPushMultiRequest.h"
#include "StreamMessageBootPushMultiResponse.h"
#include "StreamMessageBootPushMultiRegister.h"
#include "StreamMessageBootPushMultiDeregister.h"

class CStreamClientPushMulti : public CStreamClient
{
private:
	enum EClientState
	{
		STOPPED = 0,
		MCAST_STREAM_FIRST = 10,
		MCAST_STREAM_BUFFERING = 11,
		MCAST_STREAM_PLAY = 12,
		MCAST_STREAM_WAIT = 13,
		UCAST_BOOTSTRAP_REQUEST = 20,
		UCAST_BOOTSTRAP_RESPONSE = 21,
		UCAST_JOIN_REQUEST = 22,
		UCAST_STREAM_FIRST = 23,
		UCAST_STREAM_BUFFERING = 24,
		UCAST_STREAM_PLAY = 25,
		UCAST_STREAM_WAIT = 26,
		UCAST_STREAM_RESET = 27
	};

	enum ERegistrationState
	{
		NOT_REGISTERED = 0,
		REGISTERED = 1
	};

	enum ELayerState
	{
		LAYER_IDLE = 0,
		LAYER_BUFFERING = 1,
		LAYER_READY = 2
	};

	class CLayersInfo
	{
	private:
		__uint32	numBuffering;
		__uint32	numReady;

		__uint32	numRegistered;

	public:
		CLayersInfo();
		virtual ~CLayersInfo() { }

		inline __uint32	NumBuffering() { return this->numBuffering; }
		inline __uint32 NumReady() { return this->numReady; }
		inline __uint32	NumRegistered() { return this->numRegistered; }
		inline __uint32 NumTotal() { return this->numBuffering + this->numReady; }

		inline void		IdleToBuffering() { this->numBuffering++; }
		inline void		BufferingToReady() { assert(this->numBuffering); this->numBuffering--; this->numReady++; }
		inline void		ReadyToBuffering() { assert(this->numReady); this->numReady--; this->numBuffering++; }

		inline void		Register() { this->numRegistered++; }
		inline void		Deregister() { assert(this->numRegistered); this->numRegistered--; }

		inline void		Reset() { this->numBuffering = 0; this->numReady = 0; this->numRegistered = 0; }
	};

	class CLayer : public CTimerInfo
	{
	private:
		// Layer
		__uint32						layer;
		CLayersInfo*					info;
		CChannel*						channel;

		// State
		ELayerState						stateLayer;
		ERegistrationState				stateRegister;
		
		// Stream
		void							(CStreamClientPushMulti::*process)(CStreamFrame frame, __uint32 layer);	// Layer processing
		set<CAddress>					neighbors;
		set<CAddress>					receivers;
		CAddress						sender;
		
		CTimer<CStreamClientPushMulti>*	timerRegister;		// Register timer
		CTimer<CStreamClientPushMulti>*	timerBuffer;		// Buffering timer

	public:
		CLayer(
			__uint32				layer,
			CSimHandler*			sim,
			CStreamClientPushMulti*	client,
			CLayersInfo*			info,
			void					(CStreamClientPushMulti::*timerBootHandler)(CTimerInfo*),
			void					(CStreamClientPushMulti::*timerRegisterHandler)(CTimerInfo*),
			void					(CStreamClientPushMulti::*timerBufferHandler)(CTimerInfo*)
			);
		virtual ~CLayer();

		inline __uint32							Layer() { return this->layer; }

		inline CTimer<CStreamClientPushMulti>*	TimerRegister() { return this->timerRegister; }
		inline CTimer<CStreamClientPushMulti>*	TimerBuffer() { return this->timerBuffer; }

		inline ELayerState						StateLayer() { return this->stateLayer; }
		inline ERegistrationState				StateRegister() { return this->stateRegister; }

		void									Buffering();
		void									Ready();

		inline void								Register() { this->stateRegister = REGISTERED; this->info->Register(); }
		inline void								Deregister() { this->stateRegister = NOT_REGISTERED; this->info->Deregister(); }

		inline set<CAddress>*					Neighbors() { return &this->neighbors; }
		inline set<CAddress>*					Receivers() { return &this->receivers; }
		inline CAddress							Sender() { return this->sender; }
		inline void								Sender(CAddress sender) { this->sender = sender; }

		void									Start(CChannel* channel);
		void									Stop();

		inline void								(CStreamClientPushMulti::*ProcessFrame())(CStreamFrame frame, __uint32 layer) { return this->process; }
		inline void								Process(void (CStreamClientPushMulti::*process)(CStreamFrame frame, __uint32 layer)) { this->process = process; }
	};

	// State
	EClientState					stateClient;

	// Simulator
	CInfoPushMulti*					info;
	CAddress						address;

	// Channel
	CChannel*						channel;

	// Buffer
	CStreamBuffer*					bufferMcast;
	CStreamBufferMulti*				bufferUcast;

	// Session
	void							(CStreamClientPushMulti::*processFrame)(CStreamFrame frame);	// Client processing
	__time							sessionTimeLastStart;
	__time							sessionTimeLastStop;
	__uint32						sessionFrameLastStart;
	__uint32						sessionFrameLastStop;

	// Unicast session
	__bitrate						bw;				// Uplink bandwidth
	__bitrate						bwUsed;			// Used uplink bandwidth

	CLayer**						layers;
	CLayersInfo						layersInfo;

	// Multicast session
	__uint32						entry;			// Multicast interface (0 - as host have only one interface)

	// Timer
	CTimer<CStreamClientPushMulti>*	timerPlayMcast;	// Playback timer (multicast channels)
	CTimer<CStreamClientPushMulti>*	timerPlayUcast; // Playback timer (unicast channels)
	CTimer<CStreamClientPushMulti>*	timerBoot;		// Bootstrap timer (global - all layers)

	__uint32						timerBufferChannel;

	// Delegates - receive 
	Delegate2<CStreamClientPushMulti, void, CAddress, CStreamFrame>*
									delegateRecvFrameMcast;
	Delegate2<CStreamClientPushMulti, void, CAddress, CStreamFrame>*
									delegateRecvFrameUcast;

	IDelegate2<void, CAddress, CPacketStream*>*
									delegateSendStream;
	IDelegate2<void, CAddress, CStreamMessage*>*
									delegateSendMessage;

	IDelegate2<void, CAddress, __uint32>*
									delegateIgmpJoin;
	IDelegate1<void, CAddress>*		delegateIgmpLeave;


	IDelegate0<CLink*>*				delegateLink;


	// Encoder
	CStreamEncoderFrame*			encoder;

	// Decoders
	CStreamDecoder*					decoder;
	CStreamDecoderFrame*			decoderMcast;
	CStreamDecoderFrame*			decoderUcast;

	// Statistics
	__uint32						statRecvFrames;
	__uint32						statDiscardedFrames;
	__uint32						statPlayFrames;
	__uint32						statSuccessFrames[3];
	__uint32						statFailFrames[3];
	
	__uint32						statPlayFirstFrame;
	__uint32						statPlayLastFrame;

	__time							statTimeClientStart;
	__time							statTimeRecvStart;
	__time							statTimePlayStart;
	__time							statTimeFinish;
	__time							statTimeWait;

	__time							statSyncDelaySum;
	__uint32						statSyncDelayCount;

public:
	CStreamClientPushMulti(
		CSimHandler*									sim,
		CInfoPushMulti*									info,
		CAddress										address,
		IDelegate2<void, CAddress, CPacketStream*>*		delegateSendStream,
		IDelegate2<void, CAddress, CStreamMessage*>*	delegateSendMessage,
		IDelegate2<void, CAddress, __uint32>*			delegateIgmpJoin,
		IDelegate1<void, CAddress>*						delegateIgmpLeave,
		IDelegate0<CLink*>*								delegateLink,
		__uint32										entry,
		__bitrate										bw
		);
	virtual ~CStreamClientPushMulti();

	virtual void				Start(CChannel* channel);
	virtual void				Stop();

	virtual void				Recv(CAddress srcAddress, CAddress dstAddress, __uint16 srcPort, __uint16 dstPort, CPacket* packet);
	
	virtual inline __uint32		StatRecvFrames() { return this->statRecvFrames; }
	virtual inline __uint32		StatDiscardedFrames() { return this->statDiscardedFrames; }
	virtual inline __uint32		StatPlayFrames() { return this->statPlayFrames; }
	virtual inline __uint32*	StatSuccessFrames() { return this->statSuccessFrames; }
	virtual inline __uint32*	StatFailFrames() { return this->statFailFrames; }
	
	virtual inline __time		StatTimeClientStart() { return this->statTimeClientStart; }
	virtual inline __time		StatTimeRecvStart() { return this->statTimeRecvStart; }
	virtual inline __time		StatTimePlayStart() { return this->statTimePlayStart; }
	virtual inline __time		StatTimeFinish() { return this->statTimeFinish; }
	virtual inline __time		StatTimeWait() { return this->statTimeWait; }

	virtual inline __uint32		StatPlayFirstFrame() { return this->statPlayFirstFrame; }
	virtual inline __uint32		StatPlayLastFrame() { return this->statPlayLastFrame; }

	virtual inline __time		StatSyncDelay() { return this->statSyncDelayCount?(this->statSyncDelaySum / this->statSyncDelayCount):-1; }

	virtual void				Finalize();

private:
	void						StartMcast();
	void						StartUcast();

	void						StopMcast();
	void						StopUcast();

	void						RecvStream(CAddress dst, CPacketStream* packet);
	void						RecvMessage(CAddress src, CStreamMessage* message);

	void						RecvFrameMcast(CAddress src, CStreamFrame frame);

	void						ProcessFrameMcastFirst(CStreamFrame frame);
	void						ProcessFrameMcastBuffering(CStreamFrame frame);
	void						ProcessFrameMcastPlay(CStreamFrame frame);
	void						ProcessFrameMcastWait(CStreamFrame frame);

	void						RecvFrameUcast(CAddress src, CStreamFrame frame);

	void						ProcessFrameUcastFirst(CStreamFrame frame);	
	void						ProcessFrameUcastBuffering(CStreamFrame frame);
	void						ProcessFrameUcastPlay(CStreamFrame frame);
	void						ProcessFrameUcastWait(CStreamFrame frame);
	void						ProcessFrameUcastReset(CStreamFrame frame);

	void						ProcessFrameUcastLayerBuffering(CStreamFrame frame, __uint32 layer);
	void						ProcessFrameUcastLayerReady(CStreamFrame frame, __uint32 layer);

	void						RecvMessageJoin(CAddress src, CStreamMessagePushMultiJoin* message);
	void						RecvMessageLeave(CAddress src, CStreamMessagePushMultiLeave* message);
	void						RecvMessageClose(CAddress src, CStreamMessagePushMultiClose* message);
	void						RecvMessageBootResponse(CAddress src, CStreamMessageBootPushMultiResponse* message);

	void						StreamBootstrap();
	void						StreamJoin();
	void						StreamJoin(__uint32 layer);
	void						StreamLeave();
	void						StreamLeave(__uint32 layer);
	void						StreamRegister(__uint32 layer);
	void						StreamDeregister(__uint32 layer);
	void						StreamClose();
	
	void						SendStream(CStreamFrame frame, __uint32 layer);

	void						RejectJoin(CAddress src, CStreamMessagePushMultiJoin* message);

	void						TimerPlayMcast(CTimerInfo* info);
	void						TimerPlayUcast(CTimerInfo* info);
	void						TimerBoot(CTimerInfo* info);
	void						TimerRegister(CTimerInfo* info);
	void						TimerBuffer(CTimerInfo* info);
};