#include "Headers.h"
#include "StreamServerPushUcast.h"

CStreamServerPushUcast::CStreamServerPushUcast(
	CSimHandler*									sim,
	CStreamSource*									source,
	IDelegate2<void, CAddress, CPacketStream*>*		delegateSendStream,
	IDelegate2<void, CAddress, CStreamMessage*>*	delegateSendMessage
	) : CStreamServer(sim)
{
	// Set source
	this->source = source;

	// Set lower layer delegates
	this->delegateSendStream = delegateSendStream;
	this->delegateSendMessage = delegateSendMessage;

	// Create the server timer and hook it to the timer method
	this->timer = new CTimer<CStreamServerPushUcast>(
		sim,
		this,
		&CStreamServerPushUcast::Timer);
	assert_mem(this->timer);

	// Set the schedule rate (the channel fps)
	this->scheduleRate = this->source->Channel()->Fps();

	// Create the bootstrap
	this->bootstrap = new CStreamBootPush(this->source->Channel()->Address());
	assert_mem(this->bootstrap);

	// Encoder
	this->encoder = new CStreamEncoderFrame(
		this->delegateSendStream
		);
	assert_mem(this->encoder);
}

CStreamServerPushUcast::~CStreamServerPushUcast()
{
	// Timer
	delete this->timer;
	// Bootstrap
	delete this->bootstrap;
	// Encoder
	delete this->encoder;
}

void CStreamServerPushUcast::Start()
{
	// Check the server timer is stopped
	assert(!this->timer->IsSet());

	// Synchronize with the stream source
	this->source->Synchronize(this->sim->Time(), this->scheduleTime, this->scheduleFrame);

	// Set the delta frame to zero
	this->deltaFrame = 0;

	// If the schedule time is the current time, call the timer function
	if(this->scheduleTime == this->sim->Time())
		this->Timer(NULL);
	else	// Else, set the timer at the scheduled time
		this->timer->SetAt(this->scheduleTime);
}

void CStreamServerPushUcast::Stop()
{
	// If the timer is set, cancel
	if(this->timer->IsSet()) this->timer->Cancel();
}

void CStreamServerPushUcast::Timer(CTimerInfo* info)
{
	// Process server timer

	// Fetch the frame from the source
	CStreamFrame frame = this->source->Fetch(this->scheduleFrame + this->deltaFrame);

	// Encode the frame to all destinations (the destination is the receiver unicast address)
	for(set<CAddress>::iterator iter = this->receivers.begin(); iter != this->receivers.end(); iter++)
	{
		this->encoder->Encode(*iter, frame);
	}

	// Increment the delta frame
	this->deltaFrame++;

	// Set a new timer
	this->timer->SetAt(this->scheduleTime + ((__time)this->deltaFrame)/this->scheduleRate);
}

void CStreamServerPushUcast::Recv(CAddress src, CStreamMessage* message)
{	
	// Check the message
	assert(message);
	assert(message->Stream() == this->source->Channel()->Id());

	// Process the message depending on the message type
	switch(message->MessageType())
	{
	case CStreamMessage::STREAM_MESSAGE_BOOT_PUSH_REQUEST: this->RecvMessageQuery(src, type_cast<CStreamMessageBootPushRequest*>(message)); break;
	case CStreamMessage::STREAM_MESSAGE_BOOT_PUSH_REGISTER: this->RecvMessageRegister(src, type_cast<CStreamMessageBootPushRegister*>(message)); break;
	case CStreamMessage::STREAM_MESSAGE_BOOT_PUSH_DEREGISTER: this->RecvMessageDeregister(src, type_cast<CStreamMessageBootPushDeregister*>(message)); break;
	case CStreamMessage::STREAM_MESSAGE_PUSH_JOIN: this->RecvMessageJoin(src, type_cast<CStreamMessagePushJoin*>(message)); break;
	case CStreamMessage::STREAM_MESSAGE_PUSH_LEAVE: this->RecvMessageLeave(src, type_cast<CStreamMessagePushLeave*>(message)); break;
	}
}

void CStreamServerPushUcast::RecvMessageQuery(CAddress src, CStreamMessageBootPushRequest* message)
{
	// Process query
	// Reply with a response
	CStreamMessageBootPushResponse* reply;

	// Retrieve a random set of neighbors
		// Get all registered hosts
	CStreamBootPush::List* hosts = this->bootstrap->Fetch();

	// Copy all hosts locally to remove any ineligible hosts (i.e. the source hosts and descendants)
	CAddress* chosts = new CAddress[hosts->size()];

	__uint32 count = 0;
	for(CStreamBootPush::List::iterator iter = hosts->begin(); iter != hosts->end(); iter++)
	{
		if(!this->bootstrap->IsDependent(src, iter->first)) chosts[count++] = iter->first;
	}

	assert(count <= hosts->size());

	if(count <= message->Count())
	{
		// If the number of hosts is less than or equal to the requested number, put all in the reply
		reply = new CStreamMessageBootPushResponse(this->source->Channel()->Id(),count);

		for(__uint32 index = 0; index < count; index++)
			reply->Host(index) = chosts[index];
	}
	else
	{
		// If the number of hosts is greater than the requested number, choose a random subset
		reply = new CStreamMessageBootPushResponse(this->source->Channel()->Id(), message->Count());

		CShuffle shuffle(count);

		for(__uint32 index = 0; index < message->Count(); index++)
			reply->Host(index) = chosts[shuffle[index]];
	}

	delete[] chosts;

	// Send the message through the delegate
	(*this->delegateSendMessage)(src, reply);
}

void CStreamServerPushUcast::RecvMessageRegister(CAddress src, CStreamMessageBootPushRegister* message)
{
	// Register a host
	assert(message->Stream() == this->source->Channel()->Id());

	this->bootstrap->Register(message->Address(), message->Sender());
}

void CStreamServerPushUcast::RecvMessageDeregister(CAddress src, CStreamMessageBootPushDeregister* message)
{
	// Deregister a host
	assert(message->Stream() == this->source->Channel()->Id());

	this->bootstrap->Deregister(src);
}

void CStreamServerPushUcast::RecvMessageJoin(CAddress src, CStreamMessagePushJoin* message)
{
	// Process join

	// Add the source to the list of receivers for the specified layer
	this->receivers.insert(src);
}

void CStreamServerPushUcast::RecvMessageLeave(CAddress src, CStreamMessagePushLeave* message)
{
	// Process leave

	// Remove the source from the list of receivers for the specified layer
	this->receivers.erase(src);
}

void CStreamServerPushUcast::Finalize()
{
	// Call base class finalizer
	CStreamServer::Finalize();
}