#pragma once

#include "SimHandler.h"
#include "TimerBase.h"
#include "EventTimer.h"

template<class T> class CTimer : public CTimerBase
{
protected:
	CSimHandler*	sim;
	T*				object;
	void			(T::*handler)(CTimerInfo* info);
	CEventTimer*	evt;

public:
	CTimer(
		CSimHandler*	sim,
		T*				object,
		void			(T::*handler)(CTimerInfo* info) = NULL
		);

	virtual ~CTimer();

	inline bool		IsSet() { return this->evt != NULL; }

	void			SetAt(__time time, CTimerInfo* info = NULL);
	void			SetAt(__time time, void (T::*handler)(CTimerInfo* info), CTimerInfo* info = NULL);
	void			SetAfter(__time interval, CTimerInfo* info = NULL);
	void			SetAfter(__time interval, void (T::*handler)(CTimerInfo* info), CTimerInfo* info = NULL);

	void			ResetAt(__time time);
	void			ResetAfter(__time interval);

	void			Cancel();

	virtual void	Finish(CSimEvent* evt);
};

template<class T> CTimer<T>::CTimer(
	CSimHandler*	sim,
	T*				object,
	void			(T::*handler)(CTimerInfo* info)
	)
{
	assert(sim);
	assert(object);

	this->sim = sim;
	this->object = object;
	this->handler = handler;
	this->evt = NULL;
}

template<class T> CTimer<T>::~CTimer()
{
	// Cancel any pending events
	if(this->evt) this->sim->CancelEvent(this->evt->Time(), this->evt);
}

template<class T> void CTimer<T>::SetAt(__time time, CTimerInfo* info)
{
	assert(NULL == this->evt);
	assert(NULL != this->handler);

	// Schedule next event
	CEventTimer* evt = new CEventTimer(this, time, info);
	assert_mem(evt);

	this->sim->ScheduleEventAt(time, evt);

	this->evt = evt;
}

template<class T> void CTimer<T>::SetAt(__time time, void (T::*handler)(CTimerInfo* info), CTimerInfo* info)
{
	assert(NULL == this->evt);
	assert(NULL != handler);

	// Set the handler
	this->handler = handler;

	// Schedule next event
	CEventTimer* evt = new CEventTimer(this, time, info);
	assert_mem(evt);

	this->sim->ScheduleEventAt(time, evt);

	this->evt = evt;
}

template<class T> void CTimer<T>::SetAfter(__time interval, CTimerInfo* info)
{
	assert(NULL == this->evt);
	assert(NULL != this->handler);

	// Schedule next event
	CEventTimer* evt = new CEventTimer(this, this->sim->Time() + interval, info);
	assert_mem(evt);

	this->sim->ScheduleEventAt(this->sim->Time() + interval, evt);

	this->evt = evt;
}

template<class T> void CTimer<T>::SetAfter(__time interval, void (T::*handler)(CTimerInfo* info), CTimerInfo* info)
{
	assert(NULL == this->evt);
	assert(NULL != handler);

	// Set the handler
	this->handler = handler;

	// Schedule next event
	CEventTimer* evt = new CEventTimer(this, this->sim->Time() + interval, info);
	assert_mem(evt);

	this->sim->ScheduleEventAt(this->sim->Time() + interval, evt);

	this->evt = evt;
}

template<class T> void CTimer<T>::ResetAt(__time time)
{
	assert(NULL != this->evt);

	CTimerInfo* info = this->evt->Info();

	// Cancel old event
	this->sim->CancelEvent(this->evt->Time(), this->evt);

	// Schedule next event
	CEventTimer* evt = new CEventTimer(this, time, info);
	assert_mem(evt);

	this->sim->ScheduleEventAt(time, evt);

	this->evt = evt;
}

template<class T> void CTimer<T>::ResetAfter(__time interval)
{
	assert(NULL != this->evt);

	CTimerInfo* info = this->evt->Info();

	// Cancel old event
	this->sim->CancelEvent(this->evt->Time(), this->evt);

	// Schedule next event
	CEventTimer* evt = new CEventTimer(this, this->sim->Time() + interval, info);
	assert_mem(evt);

	this->sim->ScheduleEventAt(this->sim->Time() + interval, evt);

	this->evt = evt;
}

template<class T> void CTimer<T>::Cancel()
{
	// Cancel timer
	assert(this->evt);

	this->sim->CancelEvent(this->evt->Time(), this->evt);

	this->evt = NULL;
}

template<class T> void CTimer<T>::Finish(CSimEvent* evt)
{
	assert(this->evt == evt);

	// Copy the info
	CTimerInfo* info = this->evt->Info();

	// Set event to null
	this->evt = NULL;

	// Call event handler
	(this->object->*this->handler)(info);
}
