#include "Headers.h"
#include "TopoBriteRoute.h"

#define MAX_DIST		((double)1.7E+308)

CTopoBriteRoute::CTopoBriteRoute(
	__uint32		numNodes,
	__uint32		numEdges,
	CTopoBriteNode**	nodes,
	CTopoBriteEdge**	edges
	)
{
	assert(nodes);
	assert(edges);
	assert(numNodes);
	assert(numEdges);

	this->numNodes = numNodes;
	this->numEdges = numEdges;

	this->nodes = nodes;
	this->edges = edges;

	// Create the route (nodes and edges)
	this->routeNode = new int[this->numNodes*this->numNodes];
	assert(this->routeNode);

	this->routeEdge = new int[this->numNodes*this->numNodes];
	assert(this->routeNode);

	// Init the route
	for(__uint32 index = 0; index < this->numNodes*this->numNodes; index++)
	{
		this->routeNode[index] = -1;
		this->routeEdge[index] = -1;
	}

	// Distance vector
	this->distance = new __uint32[this->numNodes*this->numNodes];
	assert(this->numNodes);

	// Calculate the shortest path tree for each node
	for(__uint32 index = 0; index < this->numNodes; index++)
		this->ShortestPath(index);
}

CTopoBriteRoute::~CTopoBriteRoute()
{
	delete[] this->routeNode;
	delete[] this->routeEdge;
	delete[] this->distance;
}

void CTopoBriteRoute::ShortestPath(__uint32 node)
{
	// The shortest path from every node in the graph to "node"

	// Allocate the distance array
	double* dist = new double[this->numNodes];
	assert(dist);

	// Allocate the results array
	bool* res = new bool[this->numNodes];

	// Init
	this->ShortestPathInit(node, dist, res);
	
	for(__uint32 numRes = this->numNodes; numRes;)
	{
		// Search for the node with the lowest distance that was not previously selected
		__uint32 bestNode;
		double bestDist = MAX_DIST;
		for(__uint32 index = 0; index < this->numNodes; index++)
		{
			if(res[index])
				if(dist[index] < bestDist)
				{
					bestDist = dist[index];
					bestNode = index;
				}
		}

		// If the minimum distance is infinite, stop
		if(MAX_DIST == bestDist) break;

		// Set the node with the minimum distance as selected
		res[bestNode] = 0;

		// For each node connected to the best node
		for(__uint32 index = 0; index < this->nodes[bestNode]->Degree(); index++)
		{
			// Get the node (the node that is at the other end of the link)
			__uint32 neighbor = this->edges[this->nodes[bestNode]->Edge(index)]->OtherNode(bestNode);

			double cost = dist[bestNode] + this->edges[this->nodes[bestNode]->Edge(index)]->Cost();

			if(cost < dist[neighbor])
			{
				dist[neighbor] = cost;
				this->distance[node*this->numNodes+neighbor] = this->distance[node*this->numNodes+bestNode] + 1;

				// The next node from "neighbor" on the route to the "node" is "bestNode"
				this->routeNode[node * this->numNodes + neighbor] = bestNode;
				this->routeEdge[node * this->numNodes + neighbor] = this->nodes[bestNode]->Edge(index);
			}
		}
		
	}
	
	// Delete the distance array
	delete[] dist;

	// Delete the results array
	delete[] res;
}

void CTopoBriteRoute::ShortestPathInit(__uint32 node, double* dist, bool* res)
{
	for(__uint32 index = 0; index < this->numNodes; index++)
	{
		this->distance[node*this->numNodes+index] = 0xFFFFFFFF;
		dist[index] = MAX_DIST;
		res[index] = 1;
	}
	this->distance[node*this->numNodes+node] = 0;
	dist[node] = 0;
}

__uint32 CTopoBriteRoute::NextEdge(__uint32 node, __uint32 dst)
{
	assert(this->routeNode[dst * this->numNodes + node] >= 0);
	return this->routeEdge[dst * this->numNodes + node];
}

__uint32 CTopoBriteRoute::NextNode(__uint32 node, __uint32 dst)
{
	assert(this->routeNode[dst * this->numNodes + node] >= 0);
	return this->routeNode[dst * this->numNodes + node];
}

__uint32 CTopoBriteRoute::Distance(__uint32 src, __uint32 dst)
{
	return this->distance[src*this->numNodes+dst];
}
